from . import match_by_structured_shapes
from . import match_by_paired_parents
from . import match_by_paired_children
from . import match_unique_leafs
from . import match_number_regex
from . import match_known_paired_number_regex
from . import match_equivalent_hardcoded_leafs
from . import match_passed_hints
from . import match_by_paired_predecessors
from . import match_by_paired_prefixes